from django.db import models
from django.contrib.postgres.fields import JSONField
from django.conf import settings
from django.utils.functional import cached_property
from rest_framework.exceptions import ValidationError
from enumfields import EnumField, Enum
from arkindex_common.enums import DataImportMode
from arkindex_common.ml_tool import MLToolType
from arkindex.project.aws import S3FileModelMixin
from arkindex.project.models import IndexableModel
from arkindex.dataimport.providers import git_providers, get_provider
from arkindex.dataimport.serializers.ml_tool import MLToolTaskSerializer
from ponos.models import Workflow, State
from os import path
import shlex
import yaml
import uuid


class DataImport(IndexableModel):
    """
    A single import workflow
    """

    creator = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='imports')
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='imports')
    mode = EnumField(DataImportMode, max_length=30)
    files = models.ManyToManyField('dataimport.DataFile', related_name='imports')
    revision = models.ForeignKey(
        'dataimport.Revision', related_name='dataimports', on_delete=models.CASCADE, blank=True, null=True)
    payload = JSONField(null=True, blank=True)
    workflow = models.OneToOneField('ponos.Workflow', on_delete=models.SET_NULL, null=True, blank=True)

    class Meta:
        ordering = ['corpus', '-created']

    @property
    def state(self):
        if not self.workflow:
            return State.Unscheduled
        # This allows annotating a DataImport queryset with "last_run" and preventing duplicate SQL queries
        if hasattr(self, 'last_run'):
            # last_run may be None when there is a workflow without any tasks
            if self.last_run is None:
                return State.Unscheduled
            return self.workflow.get_state(self.last_run)
        else:
            return self.workflow.state

    def build_workflow(self):
        '''
        Create a ponos workflow with a recipe according to configuration
        '''
        ml_workflow_chunks = 1
        import_task_name = 'import'
        if self.mode == DataImportMode.Elements:
            import_task_name = 'initialisation'

        assert self.workflow is None, 'A workflow is already setup'
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')

        if self.mode == DataImportMode.Repository:
            assert self.revision is not None, \
                'A revision is required to create an import workflow from GitLab repository'
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': (f"python -m arkindex_tasks.import_git {self.revision.id}"
                                f" {self.revision.repo.credentials.token}"),
                },
            }

        elif self.mode == DataImportMode.Elements:
            assert self.payload, \
                'A payload is required to create build import workflow from Arkindex elements'
            assert self.payload.get('corpus_id'), \
                '"corpus_id" is missing in Elements dataimport payload'
            chunks = self.payload.get('chunks')
            if chunks is not None:
                assert isinstance(chunks, int) and chunks > 0, 'Chunks count should be an positive integer'
                assert chunks < 33, 'Import distribution is limited to 32 chunks'
                ml_workflow_chunks = chunks

            args = [
                'python', '-m', 'arkindex_tasks.init_elements',
                '--corpus-id', self.payload['corpus_id'],
                '--chunks-number', str(ml_workflow_chunks),
            ]

            elt_id = self.payload.get('element')
            if elt_id:
                args.append('--element-id')
                args.append(shlex.quote(elt_id))
            substring = self.payload.get('name_contains')
            if substring:
                args.append('--name-contains')
                args.append(shlex.quote(substring))
            elt_type = self.payload.get('elt_type')
            if elt_type:
                args.append('--type')
                args.append(shlex.quote(elt_type))
            best_class = self.payload.get('best_class')
            if best_class is not None:
                args.append('--best-class')
                args.append(shlex.quote(str(best_class)))
            elements = self.payload.get('elements')
            if elements:
                args.append('--dataimport-id')
                args.append(str(self.id))

            # Do not retrieve elements children in case of thumbnails generation only
            thumbnails = self.payload.get('thumbnails')
            if thumbnails and len(self.ml_tools) == 0:
                args.append('--skip-children')

            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': ' '.join(args),
                }
            }

        elif self.mode == DataImportMode.IIIF:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python -m arkindex_tasks.import_iiif.dataimport {}'.format(self.id),
                },
            }

        else:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python -m arkindex_tasks.import_files {}'.format(self.id),
                },
            }

        elts_chunk_files = ['elements.json']
        if ml_workflow_chunks > 1:
            elts_chunk_files = ['elements_chunk_{}.json'.format(n) for n in range(1, ml_workflow_chunks + 1)]

        for index, elts_chunk in enumerate(elts_chunk_files, start=1):
            # Add a name suffix if task is distributed
            task_suffix = ''
            if len(elts_chunk_files) > 1:
                task_suffix = '_{}'.format(index)

            elements_path = shlex.quote(path.join('/data', import_task_name, elts_chunk))

            # Generate thumbnails directly after import step
            if self.mode != DataImportMode.Elements or self.payload and self.payload.get('thumbnails'):
                tasks['thumbnails{}'.format(task_suffix)] = {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails {}'.format(elements_path),
                    'parents': [import_task_name],
                }

            # Write tools recipes
            for tool in self.ml_tools:
                # Override task elements file path
                task = tool.get_task_recipe(elements_path=elements_path)

                # Handle parent ML tools and default to the import task
                task['parents'] = ['ml_{}{}'.format(parent.slug, task_suffix) for parent in tool.parents]
                if not task['parents']:
                    task['parents'] = [import_task_name]

                tasks['ml_{}{}'.format(tool.slug, task_suffix)] = dict(task)

        # Build the workflow in db
        recipe = settings.PONOS_RECIPE.copy()
        recipe['tasks'] = tasks
        return Workflow.objects.create(recipe=yaml.dump(recipe))

    def start(self):
        self.workflow = self.build_workflow()
        self.save()

    def retry(self):
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')
        if self.workflow:
            self.workflow.retry()
        else:
            self.start()

    @cached_property
    def repo(self):
        if self.revision:
            return self.revision.repo
        elif 'repo_id' in self.payload:
            return Repository.objects.get(id=self.payload['repo_id'])

    @cached_property
    def sha(self):
        return self.revision.hash if self.revision else self.payload.get('sha')

    @cached_property
    def ml_tools(self):
        if not self.payload or not self.payload.get('ml_tools'):
            return []
        serializer = MLToolTaskSerializer(data=self.payload['ml_tools'], many=True)
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data


MAX_FILENAME_LENGTH = 255
"""
The maximum length for DataFile filenames.
Used in the file upload API because accessing the max_length attribute on a field
inside a Django model is unnecessarily complex.
"""


class DataFile(S3FileModelMixin):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=MAX_FILENAME_LENGTH)
    size = models.PositiveIntegerField(help_text='file size in bytes')
    content_type = models.CharField(max_length=50)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='files')

    s3_bucket = settings.AWS_STAGING_BUCKET

    @cached_property
    def s3_key(self):
        return str(self.id)

    class Meta:
        unique_together = (('corpus', 'hash'), )
        ordering = ['corpus', 'name']

    def perform_check(self, save=True, raise_exc=False):
        """
        Check Datafile's existence and update size, content_type from S3
        """
        self.check_hash(save=save, raise_exc=raise_exc)

        self.size = self.s3_object.content_length
        self.content_type = self.s3_object.content_type
        if save:
            self.save()


class Repository(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    url = models.URLField(unique=True)
    hook_token = models.CharField(max_length=250, unique=True)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='repos')
    credentials = models.ForeignKey(
        'users.OAuthCredentials', on_delete=models.SET_NULL, related_name='repos', blank=True, null=True)
    provider_name = models.CharField(
        max_length=50,
        choices=[(p.__name__, p.display_name) for p in git_providers],
    )
    git_ref_revisions = models.ManyToManyField('dataimport.Revision', through='dataimport.GitRef')

    class Meta:
        verbose_name_plural = 'repositories'

    @property
    def provider_class(self):
        return get_provider(self.provider_name)

    @property
    def provider(self):
        return self.provider_class(credentials=self.credentials)

    @property
    def enabled(self):
        return self.credentials is not None and self.provider_class is not None

    def __str__(self):
        return f'{self.url}'


class GitRefType(Enum):
    Branch = 'branch'
    Tag = 'tag'


class GitRef(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    type = EnumField(GitRefType)
    name = models.CharField(max_length=250)
    repository = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='refs')
    revision = models.ForeignKey('dataimport.Revision', on_delete=models.CASCADE, related_name='refs')

    class Meta:
        unique_together = (('name', 'repository'), )


class RevisionState(Enum):
    Created = 'created'
    Processing = 'processing'
    Available = 'available'
    Error = 'error'


class Revision(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    repo = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='revisions')
    hash = models.CharField(max_length=50)
    message = models.TextField()
    author = models.CharField(max_length=50)
    state = EnumField(RevisionState, default=RevisionState.Created)

    class Meta:
        unique_together = (('repo', 'hash'), )

    @property
    def commit_url(self):
        return '{}/commit/{}'.format(self.repo.url.rstrip('/'), self.hash)

    def start_import(self):
        dataimport = self.dataimports.create(
            creator=self.repo.credentials.user,
            corpus=self.repo.corpus,
            mode=DataImportMode.Repository,
            payload={
                'repo_id': str(self.repo.id),
                'sha': self.hash,
            }
        )
        dataimport.start()
        return dataimport

    def __str__(self):
        return '{} "{}" by {}'.format(self.hash[:8], self.message.splitlines()[0], self.author)


class Worker(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    slug = models.CharField(max_length=100)
    type = EnumField(MLToolType, max_length=50)
    repository = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='workers')

    class Meta:
        unique_together = (('slug', 'repository'),)

    def __str__(self):
        return f'{self.name}'


class WorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    worker = models.ForeignKey('dataimport.Worker', on_delete=models.CASCADE)
    revision = models.ForeignKey('dataimport.Revision', on_delete=models.CASCADE)
    configuration = JSONField()

    class Meta:
        unique_together = (('worker', 'revision'),)

    def __str__(self):
        return f'{self.worker} for revision {self.revision}'
