from collections import defaultdict
from django.conf import settings
from django.db.models import Q, Count
from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.mixins import CreateModelMixin
from rest_framework.generics import GenericAPIView, ListAPIView, CreateAPIView, UpdateAPIView, RetrieveDestroyAPIView
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView
from arkindex.documents.models import \
    Corpus, Element, DataSource, Classification, ClassificationState, Transcription, Entity, Right, MLClass
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import EntityType
from arkindex.documents.serializers.ml import (
    ClassificationsSerializer, ClassificationCreateSerializer, ClassificationSerializer,
    TranscriptionsSerializer, TranscriptionCreateSerializer, DataSourceStatsSerializer,
    ClassificationsSelectionSerializer, ClassificationMode,
    CountMLClassSerializer, MLClassSerializer
)
from arkindex.documents.pagexml import PageXmlParser
from arkindex.images.importer import build_transcriptions, save_transcriptions
from arkindex.project.filters import SafeSearchFilter
from arkindex.project.mixins import SelectionMixin, CorpusACLMixin
from arkindex.project.parsers import XMLParser
from arkindex.project.permissions import IsVerified, IsAdminUser
from arkindex.project.triggers import reindex_start, ml_results_delete
import os.path
import logging

logger = logging.getLogger(__name__)


class TranscriptionCreate(CreateAPIView):
    """
    Create a single transcription on an element
    The transcription zone is defined by the element it is attached
    """
    serializer_class = TranscriptionCreateSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'CreateTranscription',
        'tags': ['elements'],
    }

    def get_object(self):
        if not hasattr(self, 'element'):
            self.element = super().get_object()
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()
        return Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.request.user)
        )

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['element'] = self.get_object()
        return context

    def perform_create(self, serializer):
        manual_source, _ = DataSource.objects.get_or_create(
            type=MLToolType.Recognizer,
            slug='manual',
            defaults={
                'revision': '',
                'internal': False,
            }
        )
        ts = Transcription.objects.create(
            element=self.element,
            source=manual_source,
            type=serializer.validated_data['type'],
            text=serializer.validated_data['text'],
        )

        # Index in ES
        reindex_start(element=self.element, transcriptions=True, elements=True)
        return ts

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response({'id': obj.id}, status=status.HTTP_201_CREATED, headers=headers)


class TranscriptionBulk(CreateAPIView, UpdateAPIView):
    '''
    Create multiple transcriptions at once, all linked to the same page
    and to the same recognizer.
    '''
    serializer_class = TranscriptionsSerializer
    permission_classes = (IsVerified, )
    # Force DRF to ignore PATCH
    http_method_names = ['post', 'put', 'head', 'options', 'trace']
    openapi_overrides = {
        'tags': ['ml'],
        'responses': {
            '201': {
                'description': 'All transcriptions have been created.'
            }
        }
    }

    def create(self, request, *args, **kwargs):
        return self.run(request, delete=False)

    def update(self, request, *args, **kwargs):
        return self.run(request, delete=True)

    def run(self, request, delete=False):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        parent = serializer.validated_data['parent']
        source = serializer.validated_data['recognizer']
        if delete:
            # Clear all previous transcriptions from source
            parent.transcriptions.filter(source=source).delete()

        trpolygons = build_transcriptions(
            image=parent.zone.image,
            parent=parent,
            source=source,
            items=[
                {
                    'polygon': tr['polygon'],
                    'text': tr['text'],
                    'score': tr['score'],
                    'type': tr['type'],
                }
                for tr in serializer.validated_data['transcriptions']
            ],
        )

        headers = self.get_success_headers(serializer.data)
        if not trpolygons:
            return Response(None, status=status.HTTP_201_CREATED, headers=headers)

        save_transcriptions(*trpolygons)
        reindex_start(element=parent, transcriptions=True, elements=True)

        return Response(None, status=status.HTTP_201_CREATED, headers=headers)


class CorpusMLClassList(CorpusACLMixin, ListAPIView):
    """
    List available classes in a corpus with their distribution over elements
    """
    serializer_class = CountMLClassSerializer
    openapi_overrides = {
        'operationId': 'ListCorpusMLClasses',
        'description': (
            'List all available classes in a corpus with their'
            ' distribution over elements (best classes count)'
        ),
        'tags': ['ml'],
    }
    filter_backends = [SafeSearchFilter]
    search_fields = ['name']

    def get_queryset(self):
        if not self.request:
            return MLClass.objects.none()
        best_classification_filter = Q(
            # Keep non rejected best or validated classifications
            (
                (
                    Q(classifications__high_confidence=True)
                    & ~Q(classifications__state=ClassificationState.Rejected)
                )
                | Q(classifications__state=ClassificationState.Validated)
            )
        )
        return MLClass.objects.filter(corpus=self.get_corpus(self.kwargs['pk'])) \
            .annotate(nb_best=Count(
                'classifications',
                filter=best_classification_filter,
            ))


class MLClassList(ListAPIView):
    """
    List all available classes
    """
    serializer_class = MLClassSerializer
    queryset = MLClass.objects.all()
    openapi_overrides = {
        'operationId': 'ListMLClasses',
        'description': 'List available machine learning classes in all corpora.',
        'tags': ['ml'],
    }


class ClassificationCreate(CreateAPIView):
    """
    Manually add a class on an element, as a human, not as a machine learning tool.
    """
    serializer_class = ClassificationCreateSerializer
    queryset = Classification.objects.all()
    openapi_overrides = {
        'tags': ['ml'],
    }

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data_source, __ = DataSource.objects.get_or_create(
            type=MLToolType.Classifier,
            slug='manual',
            defaults={
                'revision': '',
                'internal': False,
            }
        )
        if Classification.objects.filter(
            element_id=serializer.validated_data['element'],
            source_id=data_source.id,
            ml_class_id=serializer.validated_data['ml_class']
        ).exists():
            raise ValidationError(
                {'__all__': ['A manual classification already exists for this `element` and this `class`']}
            )
        serializer.save(
            source=data_source,
            moderator=self.request.user,
            state=ClassificationState.Validated,
            high_confidence=True,
            confidence=1
        )
        headers = self.get_success_headers(serializer.data)
        response_serializer = ClassificationSerializer(instance=serializer.instance)
        return Response(response_serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class ClassificationBulk(CreateAPIView):
    """
    Create multiple classifications at once on the same element with the same classifier.
    """
    serializer_class = ClassificationsSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'tags': ['ml'],
    }

    def perform_create(self, serializer):
        parent = serializer.validated_data['parent']
        source = serializer.validated_data['classifier']
        parent.classifications.filter(source=source).delete()
        Classification.objects.bulk_create([
            Classification(
                element=parent,
                source=source,
                ml_class=cl['ml_class'],
                confidence=cl['confidence'],
                high_confidence=cl['high_confidence']
            )
            for cl in serializer.validated_data['classifications']
        ])


class ManageClassificationsSelection(SelectionMixin, CreateAPIView):
    """
    Create multiple classifications at once for a list of selected elements for a specific corpus
    """
    serializer_class = ClassificationsSelectionSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'description': 'Manage classifications for a list of selected elements for a specific corpus.',
        'tags': ['ml']
    }

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        corpus_id = serializer.validated_data['corpus_id']
        corpus = self.get_corpus(corpus_id, right=Right.Write)

        mode = serializer.validated_data['mode']
        if mode == ClassificationMode.Create:
            return self.create(corpus, request, *args, **kwargs)
        elif mode == ClassificationMode.Validate:
            elements = self.get_selection(corpus_id)
            Classification.objects.filter(
                element__in=elements,
                state=ClassificationState.Pending,
                high_confidence=True
            ).update(state=ClassificationState.Validated)
            return Response(status=status.HTTP_201_CREATED)
        raise NotImplementedError

    def create(self, corpus, request, *args, **kwargs):
        ml_class = MLClass.objects.filter(id=request.data['ml_class']).first()
        data_source, _ = DataSource.objects.get_or_create(
            type=MLToolType.Classifier,
            slug='manual',
            defaults={
                'revision': '',
                'internal': False,
            }
        )

        elements = self.get_selection(corpus.id)
        existing_element_ids = set(Classification.objects.filter(
            element_id__in=[element for element in elements],
            source_id=data_source.id,
            ml_class_id=ml_class.id
        ).values_list('element_id', flat=True))

        classifications = []
        for element in elements.exclude(id__in=existing_element_ids):
            classifications.append(Classification(
                element=element,
                ml_class=ml_class,
                source=data_source,
                moderator=self.request.user,
                state=ClassificationState.Validated,
                high_confidence=False,
                confidence=1
            ))
        Classification.objects.bulk_create(classifications)

        return Response(
            status=status.HTTP_201_CREATED
        )


class ClassificationModerationActionsMixin(GenericAPIView):
    serializer_class = ClassificationSerializer

    def get_queryset(self):
        return Classification.objects.filter(element__corpus__in=Corpus.objects.writable(self.request.user))

    def put(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.moderator = self.request.user
        instance.state = self.moderation_action
        instance.save(update_fields=['moderator', 'state'])
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)


class ClassificationValidate(ClassificationModerationActionsMixin):
    """
    Validate a classification
    """
    moderation_action = ClassificationState.Validated
    openapi_overrides = {
        'operationId': 'ValidateClassification',
        'description': 'Validate an existing classification.',
        'tags': ['ml'],
    }


class ClassificationReject(ClassificationModerationActionsMixin):
    """
    Reject a classification
    """
    moderation_action = ClassificationState.Rejected
    openapi_overrides = {
        'operationId': 'RejectClassification',
        'description': 'Reject an existing classification. Manual classifications will be deleted.',
        'tags': ['ml'],
    }

    def put(self, request, *args, **kwargs):
        instance = self.get_object()
        if instance.source.slug == 'manual':
            # Delete manual classifications upon rejection
            instance.delete()
            return Response(None, status=status.HTTP_204_NO_CONTENT)

        instance.moderator = self.request.user
        instance.state = self.moderation_action
        instance.save(update_fields=['moderator', 'state'])
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)


class PageXmlTranscriptionsImport(CreateModelMixin, APIView):
    parser_classes = (XMLParser, )
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'ImportPageXmlTranscriptions',
        'description': 'Import transcriptions into Arkindex from region data in the PAGE XML format.',
        'tags': ['ml'],
        'requestBody': {
            'required': True,
            'description': 'A PAGE XML document. TextRegion tags will be imported as Paragraph transcriptions '
                           'and TextLine tags will become Line transcriptions. '
                           'See https://github.com/PRImA-Research-Lab/PAGE-XML for more info '
                           'about the PAGE XML format.',
            'content': {
                'application/xml': {
                    'schema': {}
                }
            }
        }
    }

    def get_queryset(self):
        return Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.request.user),
            zone__isnull=False,
        )

    def get_object(self):
        """
        Since we are inheriting from APIView, because GenericAPIView would break OpenAPI,
        we have to rewrite get_object ourselves.
        """
        obj = get_object_or_404(self.get_queryset(), pk=self.kwargs['pk'])
        self.check_object_permissions(self.request, obj)
        return obj

    def post(self, request, *args, **kwargs):
        element = self.get_object()
        try:
            parser = PageXmlParser(request.data)
        except AssertionError as e:
            raise ValidationError('Could not parse PAGE XML document: {!s}'.format(e))

        transcriptions = parser.save(element)

        blocks = parser.create_blocks(transcriptions)
        blocks = parser.merge(blocks)

        path = os.path.join(settings.BASE_DIR, 'documents', 'fixtures', 'roles_transkribus.yml')
        types = {
            '_nom': EntityType.Person.value,
            '_prenom': EntityType.Person.value,
            '_lieu_residence': EntityType.Location.value,
            '_profession': EntityType.Subject.value,
        }
        entities_id = parser.create_objects(blocks, element, path, types)
        logger.info('{} blocks: {} entities created'.format(
            len(blocks),
            len(entities_id)
        ))

        reindex_start(element=element, transcriptions=True, elements=True, entities=True)

        return Response(
            status=status.HTTP_201_CREATED,
            headers=self.get_success_headers(None),
        )


class MLStatsBase(object):
    serializer_class = DataSourceStatsSerializer
    permission_classes = (IsAdminUser, )
    # Make DRF understand we return an unpaginated list, for OpenAPI schema generation
    action = 'list'
    pagination_class = None

    def get_count_querysets(self, instance):
        """
        Given an object returned by Django REST Framework's get_object,
        should return a dict mapping attribute names to querysets
        """
        if isinstance(instance, Element):
            if not instance.type.folder:
                return {
                    'transcriptions_count': Transcription.objects.filter(element_id=instance.id),
                    'entities_count': Entity.objects.filter(
                        Q(transcriptions__element_id=instance.id)
                        | Q(metadatas__element_id=instance.id)
                    ),
                    'classifications_count': Classification.objects.filter(element_id=instance.id),
                }
            # The folder AND its children
            elements = Element.objects.filter(id=instance.id).values('id').union(
                # Disable ordering here because we do not need it and it adds an extra column,
                # causing the UNION to fail
                Element.objects.get_descending(instance.id).order_by().values('id')
            )
        elif isinstance(instance, Corpus):
            elements = instance.elements.all()
        else:
            raise ValueError('Instance is not a corpus or an element: {}'.format(instance))

        return {
            'transcriptions_count': Transcription.objects.filter(element__in=elements),
            'entities_count': Entity.objects.filter(
                Q(transcriptions__element__in=elements)
                | Q(metadatas__element__in=elements)
            ),
            'classifications_count': Classification.objects.filter(element__in=elements),
        }

    def get_counts(self):
        count_querysets = self.get_count_querysets(self.get_object())

        # A dict that links source IDs to another dict holding their stats:
        # {id: {transcriptions_count: 42, …}, …}
        counts = defaultdict(dict)

        # Request statistics for each kind of ML result, grouped by source ID: lowers the need for any joins or unions
        for field_name, queryset in count_querysets.items():
            queryset = queryset.values('source_id').annotate(count=Count('id')).values_list('source_id', 'count')
            for source_id, count in queryset:
                counts[source_id][field_name] = count

        # Fetch the source IDs returned by the previous queries,
        # and set the counts as if they were usual queryset annotations.
        sources = list(DataSource.objects.filter(id__in=counts.keys()).order_by('name'))
        for source in sources:
            for name, value in counts[source.id].items():
                setattr(source, name, value)

        return sources

    def get_serializer(self, *args, **kwargs):
        # Force the serializer to work as a list, except when get_serializer is called without arguments
        # because the OpenAPI schema generator needs a normal serializer
        if args or kwargs:
            kwargs['many'] = True
        return super().get_serializer(*args, **kwargs)

    def retrieve(self, *args, **kwargs):
        serializer = self.get_serializer(self.get_counts())
        return Response(serializer.data)


class ElementMLStats(MLStatsBase, RetrieveDestroyAPIView):
    openapi_overrides = {
        'operationId': 'RetrieveElementMLStats',
        'description': 'List machine learning result sources along with their result counts for an element',
        'tags': ['ml'],
    }

    def get_queryset(self):
        return Element.objects.filter(corpus__in=Corpus.objects.readable(self.request.user)).select_related('type')

    def destroy(self, *args, **kwargs):
        ml_results_delete(element=self.get_object())
        return Response(status=status.HTTP_204_NO_CONTENT)


class CorpusMLStats(MLStatsBase, RetrieveDestroyAPIView):
    openapi_overrides = {
        'operationId': 'RetrieveCorpusMLStats',
        'description': 'List machine learning results sources along with their result counts '
                       'for all elements in a corpus',
        'tags': ['ml'],
    }

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user).only('id')

    def destroy(self, *args, **kwargs):
        ml_results_delete(corpus=self.get_object())
        return Response(status=status.HTTP_204_NO_CONTENT)
