from uuid import UUID
from django.urls import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import MetaType, TranscriptionType, EntityType
from arkindex.documents.models import \
    Element, DataSource, Transcription, Corpus, Entity, MLClass, Classification
from arkindex.images.models import ImageServer
from arkindex.project.aws import S3FileStatus
from arkindex.project.polygon import Polygon
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User


class TestElementsAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.page_type = cls.corpus.types.get(slug='page')
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.element = Element.objects.get(name='Volume 1, page 2r')
        cls.image = ImageServer.objects.local.images.create(
            path="kingdom/far/away",
            status=S3FileStatus.Checked,
            width=42,
            height=42,
        )
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.private_elt = cls.private_corpus.elements.create(type=cls.private_corpus.types.create(slug='type'))

    def setUp(self):
        self.page = self.corpus.elements.get(name='Volume 1, page 1r')
        self.folio_metadata = self.page.metadatas.get()
        self.metadata = self.page.metadatas.create(
            name='meta',
            type=MetaType.Text,
            value='text_metadata'
        )

    def test_get_element(self):
        data_source = DataSource.objects.get(slug='test')
        ml_class = MLClass.objects.create(name='text', corpus=self.corpus)
        classification = Classification.objects.create(element=self.vol, source=data_source, ml_class=ml_class)

        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.vol.id))
        self.assertEqual(data['name'], 'Volume 1')
        self.assertEqual(data['corpus']['id'], str(self.corpus.id))
        self.assertEqual(data['type'], 'volume')
        self.assertEqual(data['thumbnail_url'], self.vol.thumbnail.s3_url)
        self.assertIsNone(data['thumbnail_put_url'])
        self.assertIsNone(data['source'])

        self.assertEqual(len(data['classifications']), 1)
        cl = data['classifications'][0]
        self.assertEqual(cl['id'], str(classification.id))
        self.assertEqual(cl['source']['id'], str(classification.source.id))
        self.assertEqual(cl['ml_class']['id'], str(classification.ml_class.id))

    def test_get_element_thumbnail_folder(self):
        """
        Check getting an element only gives a thumbnail URL with folders
        """
        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['thumbnail_url'], self.vol.thumbnail.s3_url)
        self.assertIsNone(data['thumbnail_put_url'])

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNone(data['thumbnail_url'])
        self.assertIsNone(data['thumbnail_put_url'])

    def test_get_element_thumbnail_put_internal(self):
        """
        Check getting an element only gives a thumbnail PUT URL to bots for folders
        """
        user = User.objects.create_internal_user('internal@internal.fr')
        self.assertTrue(user.is_internal)
        self.client.force_login(user)

        self.assertTrue(self.vol.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['thumbnail_put_url'], self.vol.thumbnail.s3_put_url)

        self.assertFalse(self.page.type.folder)
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNone(response.json()['thumbnail_put_url'])

    def test_patch_element_unverified(self):
        """
        Check patching an element with an unverified email is not possible
        """
        user = User.objects.create_user('nope@nope.fr')
        self.assertFalse(user.verified_email)
        self.client.force_login(user)

        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have permission to perform this action.'}
        )

    def test_patch_no_write_access(self):
        # Create read_only corpus right
        self.private_corpus.corpus_right.create(user_id=self.user.id)
        self.assertTrue(self.user.verified_email)
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.private_elt.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have write access to this element.'}
        )

    def test_patch_element_no_read_access(self):
        """
        Check patching an element as anonymous user is not possible
        """
        ext_user = User.objects.create_user(email='ark@ark.net')
        ext_user.verified_email = True
        ext_user.save()
        self.client.force_login(ext_user)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.private_elt.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_element(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={'name': 'Untitled (2)'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['name'], 'Untitled (2)')
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.name, 'Untitled (2)')

    def test_patch_element_name_type_only(self):
        """
        Ensure patching an element only updates name and type, nothing else,
        and every other property is ignored
        """
        self.client.force_login(self.user)
        box_type = self.corpus.types.create(slug='xyzzy', display_name='box', folder=True)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={
                'id': 'beef',
                'name': 'New name',
                'type': 'xyzzy',
                'corpus': 'something',
                'thumbnail_url': '/dev/urandom',
                'thumbnail_put_url': '/dev/null',
                'zone': 51,
                'metadata': [],
                'classifications': [],
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['name'], 'New name')
        self.assertEqual(response.json()['type'], 'xyzzy')
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.name, 'New name')
        self.assertEqual(self.vol.type, box_type)
        self.assertNotEqual(self.vol.id, 'beef')
        self.assertNotEqual(self.vol.corpus_id, 'something')

    def test_patch_element_polygon(self):
        self.client.force_login(self.user)
        self.assertIsNotNone(self.element.zone)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.element.id)}),
            data={
                'polygon': [[10, 20], [10, 30], [30, 30], [30, 20], [10, 20]]
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['name'], self.element.name)
        self.assertEqual(response.json()['type'], self.element.type.slug)
        self.element.refresh_from_db()
        self.assertEqual(self.element.zone.polygon.x, 10)
        self.assertEqual(self.element.zone.polygon.y, 20)
        self.assertEqual(self.element.zone.polygon.width, 20)
        self.assertEqual(self.element.zone.polygon.height, 10)

    def test_patch_element_polygon_without_image(self):
        self.client.force_login(self.user)
        self.vol.zone = None
        self.vol.save()
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={
                'polygon': [[10, 20], [10, 30], [30, 30], [30, 20], [10, 20]]
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'image': ['Image is required when defining a polygon on an element without an existing zone']
        })

    def test_patch_element_image_preserve_polygon(self):
        self.client.force_login(self.user)
        self.assertNotEqual(self.element.zone.image, self.image)
        expected_polygon = self.element.zone.polygon.copy()
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.element.id)}),
            data={
                'image': str(self.image.id),
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.element.refresh_from_db()
        self.assertEqual(self.element.zone.image, self.image)
        self.assertEqual(self.element.zone.polygon, expected_polygon)

    def test_patch_element_new_zone(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.vol.zone)
        response = self.client.patch(
            reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}),
            data={
                'image': str(self.image.id),
                'polygon': [[10, 20], [10, 30], [30, 30], [30, 20], [10, 20]],
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.vol.refresh_from_db()
        self.assertEqual(self.vol.zone.image, self.image)
        self.assertEqual(self.vol.zone.polygon.x, 10)
        self.assertEqual(self.vol.zone.polygon.y, 20)
        self.assertEqual(self.vol.zone.polygon.width, 20)
        self.assertEqual(self.vol.zone.polygon.height, 10)

    def make_create_request(self, name='default', corpus=None, elt_type='volume', **options):
        request = {
            'path': reverse('api:elements-create'),
            'data': {
                'type': elt_type,
                'name': name,
                'corpus': corpus or str(self.corpus.id),
            },
            'format': 'json',
        }
        # Optional parameters
        request['data'].update(options)
        return request

    def test_create_element_not_verified(self):
        self.assertEqual(
            self.client.post(**self.make_create_request('no')).status_code,
            status.HTTP_403_FORBIDDEN
        )

    def test_create_element_volume(self):
        # Create a Volume
        self.client.force_login(self.user)
        request = self.make_create_request('my new volume')
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        data = response.json()
        self.assertIn('id', data)
        volume = Element.objects.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(volume.id),
                'corpus': str(volume.corpus.id),
                'name': volume.name,
                'type': volume.type.slug,
                'parent': None,
                'source': None,
            }
        )
        self.assertEqual(volume.name, 'my new volume')
        self.assertEqual(volume.type, self.volume_type)

    def test_create_volume_page(self):
        # Add a page with an Image to an existing volume
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams',
            image=str(self.image.id),
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'corpus': str(page.corpus.id),
                'name': page.name,
                'type': page.type.slug,
                'parent': None,
                'source': None,
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams')
        self.assertEqual(page.type, self.page_type)

    def test_create_volume_act(self):
        # Create an volume child of type Act
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            name='Castle story',
            elt_type='act'
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)

    def test_create_element_source(self):
        # Create an element with a source
        self.client.force_login(self.user)
        source = DataSource.objects.create(
            type=MLToolType.DLAnalyser,
            slug='fairy_tale_detector',
            internal=False,
        )
        request = self.make_create_request(
            name='Castle story',
            elt_type='act',
            source='fairy_tale_detector',
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.source, source)

    def test_create_element_polygon(self):
        # Create an element with a polygon to an existing volume
        polygon = [[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]]
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams again',
            image=str(self.image.id),
            polygon=polygon
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'corpus': str(page.corpus.id),
                'name': page.name,
                'type': page.type.slug,
                'parent': None,
                'source': None,
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams again')
        self.assertEqual(page.type, self.page_type)

    def test_create_parent_different_corpus(self):
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.corpus_right.create(user=self.user, can_write=True)
        new_corpus.types.create(slug='act')
        request = self.make_create_request(
            parent=str(self.vol.id),
            corpus=str(new_corpus.id),
            elt_type='act'
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'corpus': ['Parent and child must be in the same corpus']
        })

    def test_create_filter_type(self):
        """
        Ensure types are filtered by corpus
        """
        self.assertTrue(self.corpus.types.filter(slug='volume').exists())
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.corpus_right.create(user=self.user, can_write=True)
        new_vol_type = new_corpus.types.create(slug='volume', display_name='Volume', folder=True)
        request = self.make_create_request(
            corpus=str(new_corpus.id),
            elt_type='volume',
            name='something',
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_vol = new_corpus.elements.get()
        self.assertEqual(new_vol.type, new_vol_type)
        self.assertEqual(new_vol.type.corpus, new_corpus)

    def test_create_wrong_type(self):
        self.assertFalse(self.corpus.types.filter(slug='kartoffelsalad').exists())
        self.client.force_login(self.user)
        request = self.make_create_request(
            corpus=str(self.corpus.id),
            elt_type='kartoffelsalad',
            name='something',
        )
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ["Type 'kartoffelsalad' not found in corpus Unit Tests"],
        })

    def test_create_image_not_checked(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.status = S3FileStatus.Error
        self.image.save()
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertEqual(element.zone.image.status, S3FileStatus.Error)

    def test_create_image_zero_dimensions(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.width = 0
        self.image.height = 0
        self.image.status = S3FileStatus.Unchecked
        self.image.save()
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'image': ['This image does not have valid dimensions.']})

    def test_element_destroy_verified_user(self):
        response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_destroy_acl(self):
        self.private_corpus.corpus_right.create(user_id=self.user.id, can_write=True, can_admin=False)
        self.client.force_login(self.user)
        castle_story = self.private_corpus.elements.create(
            type=self.volume_type,
            name='Castle story'
        )
        response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this element.'}
        )

    def test_element_destroy(self):
        self.client.force_login(self.user)
        castle_story = self.corpus.elements.create(
            type=self.volume_type,
            name='Castle story'
        )
        self.assertEqual(self.corpus.elements.filter(id=castle_story.id).exists(), True)
        response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(castle_story.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(self.corpus.elements.filter(id=castle_story.id).exists(), False)

    def test_non_empty_folder(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:element-retrieve', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element': "Element '{}' is linked to 14 children elements".format(self.vol.id)}
        )

    def test_get_transcriptions(self):
        self.page.transcriptions.all().delete()
        transcriptions = []
        for i in range(10):
            zone, _ = self.page.zone.image.zones.get_or_create(polygon=Polygon.from_coords(0, 0, i + 1, i + 1))
            transcriptions.append(self.page.transcriptions.create(
                source_id=DataSource.objects.get(slug='test').id,
                type=TranscriptionType.Word,
                zone=zone,
            ))
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 10)
        self.assertEqual(len(transcriptions), 10)
        # Wrong transcription type
        url = reverse('api:element-transcriptions', kwargs={'pk': str(self.page.id)})
        url += '?type=potato'
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_elements_filter_name(self):
        nameSelected = 'ol'
        response = self.client.get(reverse('api:elements'), data={'name': nameSelected})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertIn(nameSelected, result['name'])

    def test_list_elements_wrong_corpus_filter(self):
        nameSelected = 'not_a_corpus_id'
        response = self.client.get(reverse('api:elements'), data={'corpus': nameSelected})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'corpus': ['Not a valid uuid']}
        )

    def test_list_elements_corpus_uuid4(self):
        """
        Ensure the corpus filter allows any UUID, not just version 4 UUIDs.
        Using the version=x argument on uuid.UUID causes the version and variant bits
        of a UUID to be changed on instanciation; since we generate v4 RFC 4122-compliant UUIDs,
        this would work in most cases - but we could have any UUID version or variant,
        generated by some other tool, which would cause strange 404 errors.
        """
        str_id = '12345678-1234-5678-1234-567812345678'
        self.assertNotEqual(
            UUID(str_id),
            UUID(str_id, version=4),
        )
        corpus = Corpus.objects.create(id=UUID(str_id), name='A corpus', public=True)
        element_type = corpus.types.create(slug='element', display_name='Generic element')
        element = corpus.elements.create(type=element_type, name='Something')

        response = self.client.get(reverse('api:elements'), data={'corpus': str_id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(element.id))
        self.assertEqual(data['results'][0]['corpus']['id'], str_id)

    def test_list_elements_filter_type(self):
        response = self.client.get(reverse('api:elements'), data={'type': 'act'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertEqual(result['type'], 'act')

    def test_list_element_metadata(self):
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['metadata']
        self.assertListEqual(results, [
            {
                'dates': [],
                'id': str(self.folio_metadata.id),
                'name': 'folio',
                'revision': None,
                'type': 'text',
                'value': '1r',
            },
            {
                'id': str(self.metadata.id),
                'name': str(self.metadata.name),
                'type': self.metadata.type.value,
                'value': self.metadata.value,
                'dates': [],
                'revision': None
            },
        ])

    def test_list_element_metadata_with_entity(self):
        # Link an entity to metadata, db requests number should remains same
        entity = Entity.objects.create(
            type=EntityType.Person,
            corpus=self.corpus,
            name='Marc',
            source_id=DataSource.objects.get(slug='test').id
        )
        self.metadata.entity = entity
        self.metadata.save()
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['metadata']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'folio')

    def test_list_element_metadata_wrong_acl(self):
        self.page.corpus = self.private_corpus
        self.page.save()
        response = self.client.get(reverse('api:element-retrieve', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_element_parents(self):
        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Volume 1', ]
        )

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?folder=false')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(self.page.id)}) + '?type=act')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        surface = Element.objects.get(name='Surface A')
        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(surface.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

        response = self.client.get(reverse('api:elements-parents', kwargs={'pk': str(surface.id)}) + '?recursive')
        self.assertEqual(len(response.json()['results']), 2)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            ['Act 1', 'Volume 1'],
        )

        response = self.client.get(
            reverse('api:elements-parents', kwargs={'pk': str(surface.id)})
            + '?recursive&name=aCT'
        )
        self.assertEqual(len(response.json()['results']), 1)
        self.assertEqual(response.json()['results'][0]['name'], 'Act 1')

    def test_element_children(self):
        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.page.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 8)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page'},
        )

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=surface'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Surface A',
                'Surface B',
                'Surface C',
                'Surface D',
                'Surface E',
                'Surface F'
            ],
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?recursive')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 14)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page', 'surface'},
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?type=page')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r'
            ]
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?name=2')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Act 2',
                'Volume 1, page 2r',
            ]
        )

        nested_volume = self.corpus.elements.create(
            type=self.vol.type,
            name='Nested volume',
        )
        nested_volume.add_parent(self.vol)

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?folder')
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "zone": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "best_classes": None,
                "children_count": None,
                "transcription_sources_count": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
                "thumbnail_put_url": None,
            }
        ])

    def test_element_children_ordering(self):
        expected = [
            'Act 4',
            'Act 5',
            'Act 1',
            'Act 2',
            'Act 3'
        ]
        for ordering, name in enumerate(expected):
            Element.objects.get(type__slug='act', name=name).paths.update(ordering=ordering)

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=act'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            expected,
        )

    def test_list_with_children_count(self):
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:elements'), data={'with_children_count': True})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['children_count'] for element in response.json()['results']},
            {
                'Act 1': 1,
                'Act 2': 2,
                'Act 3': 1,
                'Act 4': 1,
                'Act 5': 1,
                'Surface A': 0,
                'Surface B': 0,
                'Surface C': 0,
                'Surface D': 0,
                'Surface E': 0,
                'Surface F': 0,
                'Volume 1': 8,
                'Volume 1, page 1r': 0,
                'Volume 1, page 1v': 0,
                'Volume 1, page 2r': 0,
                'Volume 2': 3,
                'Volume 2, page 1r': 0,
                'Volume 2, page 1v': 0,
                'Volume 2, page 2r': 0,
            }
        )

    def test_children_with_children_count(self):
        with self.assertNumQueries(9):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'with_children_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['children_count'] for element in response.json()['results']},
            {
                'Act 1': 1,
                'Act 2': 2,
                'Act 3': 1,
                'Act 4': 1,
                'Act 5': 1,
                'Volume 1, page 1r': 0,
                'Volume 1, page 1v': 0,
                'Volume 1, page 2r': 0,
            }
        )

    def test_parents_with_children_count(self):
        surface = Element.objects.get(name='Surface A')
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(surface.id)}),
                data={'recursive': True, 'with_children_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['children_count'] for element in response.json()['results']},
            {
                'Act 1': 1,
                'Volume 1': 8,
            }
        )

    def test_with_children_count_unique(self):
        r"""
        Ensure with_children_count only counts children elements once, not even when ElementPaths are duplicated.
        Counts may have duplicates as soon as an element's parent has multiple parents:

        A   B    Element | Path
         \ /     C       | A
          C      C       | B
          |      D       | A C
          D      D       | B C

        Here, requesting the parents of D with children counts should return C with 1 child,
        although there are two ElementPaths whose last ID the path is C's.
        """
        act = Element.objects.get(name='Act 1')
        act.add_parent(Element.objects.get(name='Volume 2'))
        surface = Element.objects.get(name='Surface A')
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(surface.id)}),
                data={'with_children_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['children_count'] for element in response.json()['results']},
            {'Act 1': 1},
        )

    def test_children_with_transcription_sources_count(self):
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'with_transcription_sources_count': True}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['transcription_sources_count'] for element in response.json()['results']},
            {
                'Act 1': 0,
                'Act 2': 0,
                'Act 3': 0,
                'Act 4': 0,
                'Act 5': 0,
                'Volume 1, page 1r': 1,
                'Volume 1, page 1v': 1,
                'Volume 1, page 2r': 1,
            }
        )

    def test_parent_transcription_sources_count_multiple(self):
        surface = Element.objects.get(name='Surface A')
        parent = Element.objects.get(name='Act 1')
        # Add transcriptions from two different sources on the parent
        sources = [
            DataSource.objects.get(slug='test'),
            DataSource.objects.create(
                type=MLToolType.Recognizer,
                slug='source2',
                name='reco2',
                revision='x',
                internal=False,
            )
        ]
        parent.transcriptions.bulk_create([
            Transcription(element=parent, zone=surface.zone, source=s)
            for s in sources
        ])
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:elements-parents', kwargs={'pk': str(surface.id)}),
                data={'with_transcription_sources_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['transcription_sources_count'] for element in response.json()['results']},
            {
                'Act 1': 2,
            }
        )
