from unittest.mock import patch, call, MagicMock
from django.test import override_settings
from botocore.exceptions import ClientError
from arkindex.project.tests import FixtureTestCase
from arkindex.project.aws import S3FileStatus
from arkindex.images.models import ImageServer


@override_settings(LOCAL_IMAGESERVER_ID=1)
class TestImage(FixtureTestCase):

    def test_path_not_encoded(self):
        """
        Assert that an image's path is not URL encoded when created
        This can yield to images with URLs such as http://.../test%34test to be encoded as http://.../test%2534test
        """
        img = self.imgsrv.images.create(path='!#%:+/*')
        img.refresh_from_db()
        self.assertEqual(img.path, '!#%:+/*')
        self.assertEqual(img.url, 'http://server/!#%:+/*')

    def test_build_url(self):
        self.assertEqual(self.imgsrv.build_url('some/path'), 'http://server/some/path')
        self.assertEqual(self.imgsrv.build_url('/some/path'), 'http://server/some/path')
        # Should handle paths that look like an absolute URL but aren't
        self.assertEqual(self.imgsrv.build_url('ark:/some/path'), 'http://server/ark:/some/path')

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_s3_url(self, presigned_url_mock):
        presigned_url_mock.return_value = 'http://somewhere'
        img = ImageServer.objects.local.images.create(path='abcd')
        self.assertEqual(img.s3_url, 'http://somewhere')
        self.assertEqual(presigned_url_mock.call_count, 1)
        self.assertEqual(presigned_url_mock.call_args, call(
            'get_object',
            Params={'Bucket': 'iiif', 'Key': 'abcd'},
        ))

    @patch('arkindex.project.aws.session.client')
    @patch('arkindex.project.aws.Config')
    def test_s3_get_url_different_region(self, config_mock, s3_client_mock):
        img = ImageServer.objects.create(s3_region='middle-earth-1').images.create(path='abcd')
        config_mock.return_value = 'config'
        img.s3_object = MagicMock()
        img.s3_url
        self.assertEqual(s3_client_mock.return_value.generate_presigned_url.call_count, 1)
        self.assertEqual(
            s3_client_mock.call_args,
            call(
                's3',
                region_name='middle-earth-1',
                config='config'
            )
        )

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_s3_put_url(self, presigned_url_mock):
        presigned_url_mock.return_value = 'http://somewhere'
        img = ImageServer.objects.local.images.create(path='abcd')
        self.assertEqual(img.s3_put_url, 'http://somewhere')
        self.assertEqual(presigned_url_mock.call_count, 1)
        self.assertEqual(presigned_url_mock.call_args, call(
            'put_object',
            Params={'Bucket': 'iiif', 'Key': 'abcd'},
        ))

    @patch('arkindex.project.aws.s3')
    def test_check_hash(self, s3_mock):
        s3_mock.Object().e_tag = '"beef"'
        img = ImageServer.objects.local.images.create(path='abcd', hash='beef', width=10, height=10)
        self.assertEqual(img.status, S3FileStatus.Unchecked)
        img.check_hash(save=False)
        img.refresh_from_db()
        self.assertEqual(img.status, S3FileStatus.Unchecked)

    @patch('arkindex.project.aws.s3')
    def test_check_hash_save(self, s3_mock):
        s3_mock.Object().e_tag = '"beef"'
        img = ImageServer.objects.local.images.create(path='abcd', hash='beef', width=10, height=10)
        img.check_hash(save=True)
        img.refresh_from_db()
        self.assertEqual(img.status, S3FileStatus.Checked)

    @patch('arkindex.project.aws.s3')
    def test_check_hash_missing(self, s3_mock):
        img = ImageServer.objects.local.images.create(path='abcd')
        self.assertEqual(img.status, S3FileStatus.Unchecked)
        with self.assertRaisesRegex(AssertionError, 'no hash'):
            img.check_hash()

    @patch('arkindex.project.aws.s3')
    def test_check_hash_not_found(self, s3_mock):
        s3_mock.Object().load.side_effect = ClientError({'Error': {'Code': '404'}}, 'head_object')
        img = ImageServer.objects.local.images.create(path='abcd', hash='beef')
        self.assertEqual(img.status, S3FileStatus.Unchecked)
        with self.assertRaisesRegex(AssertionError, 'No file content, assert file has been correctly uploaded'):
            img.check_hash()

    @patch('arkindex.project.aws.s3')
    def test_check_hash_error(self, s3_mock):
        s3_mock.Object().e_tag = '"wrong"'
        img = ImageServer.objects.local.images.create(path='abcd', hash='beef')
        self.assertEqual(img.status, S3FileStatus.Unchecked)
        img.check_hash()
        self.assertEqual(img.status, S3FileStatus.Error)

    @patch('arkindex.project.aws.s3')
    def test_check_hash_exc(self, s3_mock):
        s3_mock.Object().e_tag = '"wrong"'
        img = ImageServer.objects.local.images.create(path='abcd', hash='beef')
        self.assertEqual(img.status, S3FileStatus.Unchecked)
        with self.assertRaisesRegex(ValueError, 'hashes do not match'):
            img.check_hash(raise_exc=True, save=False)
        self.assertEqual(img.status, S3FileStatus.Error)
        img.refresh_from_db()
        self.assertEqual(img.status, S3FileStatus.Unchecked)
